from typing import List, Optional

import matplotlib as mlp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from mu_map.polar_map.get_perfusion import SEGMENTS
from mu_map.polar_map.prepare import headers
from mu_map.random_search.eval.util import color_lists

plt.rcParams.update(
    {
        "text.usetex": True,
    }
)


def absolute_percent_error(prediction: np.ndarray, target: np.ndarray) -> float:
    """
    Compute the absolute percent error (APE).

    See Chen et al. "Cross-vendor, cross-tracer, and ..." (2022).

    Parameters
    ----------
    prediction: np.array
        the predicted values
    target: np.array
        the target values

    Returns
    -------
    float
        the APE score
    """
    mean_p = prediction.mean(axis=0)
    mean_t = target.mean(axis=0)
    diff = np.absolute(mean_p - mean_t)
    return 100.0 * diff / mean_t


def same_limits(axes: List[mlp.axes.Axes]):
    """
    Adjust the limits of a all axes to be the same.

    Parameters
    ----------
    axes: List[mlp.axes.Axes]
        axes whose limits are adjusted
    """
    x_min = min(map(lambda ax: ax.get_xlim()[0], axes))
    y_min = min(map(lambda ax: ax.get_ylim()[0], axes))
    x_max = max(map(lambda ax: ax.get_xlim()[1], axes))
    y_max = max(map(lambda ax: ax.get_ylim()[1], axes))
    for ax in axs:
        ax.set_xlim((x_min, x_max))
        ax.set_ylim((y_min, y_max))


def bland_altman(
    data1: np.ndarray,
    data2: np.ndarray,
    ax: Optional[mlp.axes.Axes] = None,
    color=color_lists["printer_friendly"][0],
    color_mean=color_lists["printer_friendly"][5],
    color_std="black",
):
    """
    Create a Bland-Altman plot.

    Code is mostly taken from https://stackoverflow.com/questions/16399279/bland-altman-plot-in-python.

    Parameters
    ----------
    data1: np.ndarray
        first data array
    data2: np.ndarray
        second data array
    ax: mlp.axes.Axes, optional
        axes to plot on, a new one is created if not given
    color: str
        color of the dots
    color_mean: str
        color of the mean difference line
    color_std: str
        color for the lines at 1.96 std
    """
    ax = plt.subplot() if ax is None else ax

    data1 = data1.flatten()
    data2 = data2.flatten()

    mean = np.mean([data1, data2], axis=0)
    diff = data1 - data2  # Difference between data1 and data2
    md = np.mean(diff)  # Mean of the difference
    sd = np.std(diff, axis=0)  # Standard deviation of the difference

    ax.axhline(md, color=color_mean, linestyle="-", alpha=0.7)
    for x in [-1.96, 1.96]:
        ax.axhline(md + x * sd, color=color_std, linestyle="--", alpha=0.7)
    ax.scatter(
        mean, diff, color=color, s=25, alpha=0.7, edgecolors="black", linewidths=0.3
    )

    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.grid(axis="both", alpha=0.5, linestyle="dotted")


def plot_corellation(
    data1: np.ndarray,
    data2: np.ndarray,
    ax: Optional[mlp.axes.Axes] = None,
    color: str = color_lists["printer_friendly"][0],
    color_line: str = color_lists["printer_friendly"][5],
    v_min: Optional[float] = None,
    v_max: Optional[float] = None,
):
    """
    Visualize the correlation between two data arrays.

    Parameters
    ----------
    data1: np.ndarray
        first data array
    data2: np.ndarray
        second data array
    ax: mlp.axes.Axes, optional
        axes to plot on, a new one is created if not given
    color: str
        color of the dots
    color_line: str
        color of the line showing the ideal correlation
    v_min: float
        minimum value used for the correlation line
    v_max: float
        maximum value used for the correlation line
    """
    ax = plt.subplot() if ax is None else ax

    data1 = data1.flatten()
    data2 = data2.flatten()

    v_min = v_min if v_min is not None else min(data1.min(), data2.min())
    v_max = v_max if v_max is not None else max(data1.max(), data2.max())

    ax.scatter(
        data1,
        data2,
        c=color,
        s=25,
        alpha=0.7,
        edgecolor="black",
        linewidths=0.3,
    )
    ax.plot((v_min, v_max), (v_min, v_max), color=color_line, alpha=0.7)

    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.grid(axis="both", alpha=0.5, linestyle="dotted")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Create visualizations to compare the perfusion of different polar maps",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--perfusion_csv",
        type=str,
        required=True,
        help="csv file containing the perfusion values",
    )
    parser.add_argument(
        "--baseline",
        choices=["symbia", "stir"],
        default="stir",
        help="baseline for comparison - either CTAC by scanner software (symbia) or STIR",
    )
    parser.add_argument(
        "--plot",
        choices=["bland_altman", "correlation"],
        default="bland_altman",
        help="the type of correlation plot to create",
    )
    parser.add_argument(
        "--save",
        type=str,
        help="save the plot as an image",
    )
    args = parser.parse_args()

    args.baseline = "ct" if args.baseline == "stir" else "symbia"

    data = pd.read_csv(args.perfusion_csv)
    keys_segments = [f"segment_{i}" for i in range(1, len(SEGMENTS) + 1)]

    perfusion_ctac = data[data[headers.ac] & (data[headers.type] == args.baseline)]
    perfusion_dlac = data[data[headers.ac] & (data[headers.type] == "dl")]
    perfusion_noac = data[~data[headers.ac]]

    perfusion_ctac = perfusion_ctac[keys_segments].values
    perfusion_dlac = perfusion_dlac[keys_segments].values
    perfusion_noac = perfusion_noac[keys_segments].values

    print(f"APE:")
    fig, axs = plt.subplots(1, 2, figsize=(10, 4.5))
    for i, (perfusion, title) in enumerate(
        [(perfusion_dlac, "DLAC"), (perfusion_noac, "No AC")]
    ):
        if args.plot == "bland_altman":
            bland_altman(perfusion, perfusion_ctac, ax=axs[i])
            axs[i].set_title(title)
        else:
            v_min = min(
                perfusion_ctac.min(), perfusion_dlac.min(), perfusion_noac.min()
            )
            v_max = max(
                perfusion_ctac.max(), perfusion_dlac.max(), perfusion_noac.max()
            )

            plot_corellation(
                perfusion, perfusion_ctac, ax=axs[i], v_min=v_min, v_max=v_max
            )

            axs[i].set_xlabel(title)
            axs[i].set_ylabel("CTAC")

        _ape = absolute_percent_error(perfusion, perfusion_ctac)
        print(f" - {title:>6}: {_ape.mean():.5f}±{_ape.std():.5f}")

    same_limits(axs)
    plt.tight_layout()

    if args.save:
        plt.savefig(args.save, dpi=300)

    plt.show()
